#!/usr/bin/env python3

'''
Write a new AES key to a YubiKey, and store it into a sqlite3 database.
'''

import distutils.spawn
import os
import sys
import subprocess


def hex2modhex(string):
    l = [ ('0123456789abcdef'[i], 'cbdefghijklnrtuv'[i]) for i in range(16) ]
    modhex = ''.join(dict(l).get(chr(j), '?') for j in range(256))
    return string.translate(modhex)

def gen_random(size):
    with open('/dev/urandom','rb') as fp:
        s = fp.read(size)
    return s

def get_public(name):
    return name.rjust(8, 'q')[:8]

def main():
    if len(sys.argv) != 3:
        print('Usage: %s <name> <db.sqlite3>' % sys.argv[0])
        print('eg: %s bobama db/yubikeys.sqlite3' % sys.argv[0])
        sys.exit(0)

    name = sys.argv[1]
    db = sys.argv[2]
    aeskey = gen_random(16).hex()
    public = get_public(name).encode('utf-8').hex()
    public_m = hex2modhex(public)
    uid = gen_random(6).hex()

    cmd = [ 'ykpersonalize',
            '-1',
            '-ofixed=h:%s' % public,
            '-ouid=%s' % uid,
            '-a%s' % aeskey
    ]

    print(cmd)

    # prefer sudo over running this script with root privileges
    if os.getuid() != 0:
        if not distutils.spawn.find_executable('sudo'):
            print('Root privileges required.')
            sys.exit(1)
        cmd = [ 'sudo' ] + cmd

    try:
        ret = subprocess.call(cmd)
    except OSError as e:
        if e.errno == os.errno.ENOENT:
            print('%s: command not found.' % cmd[0])
            sys.exit(1)
        else:
            raise

    # if ret != 0:
    #     sys.exit(ret)

    cwd = os.path.dirname(os.path.realpath(__file__))
    dbconf = 'yubikeyedup-dbconf'
    subprocess.call([ dbconf, '-yk', name, db ])
    subprocess.call([ dbconf, '-ya', name, public_m, uid, aeskey, db ])
    subprocess.call([ dbconf, '-yl', db])

if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import os
import sqlite3
import sys

TABLES = [
'''yubikeys(
 nickname varchar(16) unique not null,
 publicname varchar(16) unique not null,
 created varchar(24) not null,
 internalname varchar(12) not null,
 aeskey varchar(32) not null,
 active boolean default true,
 counter integer not null default 1,
 time integer not null default 1
)''',
'''oathtokens(
 nickname varchar(16) unique not null,
 publicname varchar(12) unique not null,
 created varchar(24) not null,
 secret varchar(40) not null,
 active boolean default true,
 counter integer not null default 1
)''',
'''apikeys(
 nickname varchar(16),
 secret varchar(28),
 id integer primary key
)''',
]

def create_db(filename):
    if os.path.exists(filename):
        print('%s already exists' % filename)
        sys.exit(1)

    conn = sqlite3.connect(filename)
    c = conn.cursor()
    for table in TABLES:
        c.execute('CREATE TABLE %s' % table)
    conn.commit()
    conn.close()

def main():
    if len(sys.argv) != 2:
        print('Usage: %s <db.sqlite3>' % sys.argv[0])
        sys.exit(0)

    create_db(sys.argv[1])

if __name__ == '__main__':
    main()



index = '''<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">
<html>
  <head>
    <style type="text/css">
      form { display: table; }
      p { display: table-row; }
      label { display: table-cell; text-align: right; }
      input { display: table-cell; margin-left: 10px; }
      input[readonly] { background-color: #E0E0E0; }
    </style>
  </head>
  <body>
    Yubico Yubikeys:<br />
    <form action="/wsapi/2.0/verify" method="GET">
      <p><label for="otp">OTP</label><input type="text" name="otp" id="otp" /></p>
      <p><label for="id">API id</label><input type="text" name="id" id="id" value="1" /></p>
      <p><label for="nonce">Nonce</label><input type="text" name="nonce" id="nonce1" readonly /></p>
      <p><input type="submit" /></p>
    </form>
    <br />

    <!--OATH/HOTP tokens:<br />
    <form action="/wsapi/2.0/oathverify" method="GET">
        <p><label for="otp">OTP</label><input type="text" name="otp" id="otp" /></p>
        <p><label for="id">Public id</label><input type="text" name="publicid" id="id" /></p>
        <p><label for="id">API id</label><input type="text" name="id" id="id" value="1" /></p>
        <p><label for="nonce">Nonce</label><input type="text" name="nonce" id="nonce2" readonly /></p>
        <p><input type="submit" /></p>
    </form>-->

    <script>
      function gen_nonce(length) {
      var text = '';
      var possible = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789";
      for (var i=0; i < length; i++)
            text += possible.charAt(Math.floor(Math.random() * possible.length));
            return text;
      }
      document.getElementById('nonce1').value = gen_nonce(16);
      //document.getElementById('nonce2').value = gen_nonce(16);
    </script>
  </body>
</html>
'''



#!/usr/bin/env python3

'''
YubiServe Key Management Tool
'''

import base64
import os
import random
import re
import sqlite3
import sys
import time


def randomChars(size):
    data = bytes()
    with open('/dev/urandom','rb') as fp:
        data = fp.read(size)
    return data


def usage():
    print('Usage: %s <options> <db.sqlite3>\n' % sys.argv[0])
    print(' -ya <nickname> <publicid> <secretid> <aeskey>\tAdd a new Yubikey')
    print(' -yk <nickname>\t\t\t\t\tDelete a Yubikey')
    print(' -yd <nickname>\t\t\t\t\tDisable a Yubikey')
    print(' -ye <nickname>\t\t\t\t\tEnable a Yubikey')
    print(' -yl\t\t\t\t\t\tList all yubikeys in database\n')

    print(' -ha <nickname> <publicid> <key>\t\tAdd a new OATH token')
    print(' -hk <nickname>\t\t\t\t\tDelete a OATH token')
    print(' -hd <nickname>\t\t\t\t\tDisable a OATH token')
    print(' -he <nickname>\t\t\t\t\tEnable a OATH token')
    print(' -hl\t\t\t\t\t\tList all OATH tokens in database\n')

    print(' -aa <keyname>\t\t\t\t\tGenerate an API Key')
    print(' -ak <keyname>\t\t\t\t\tRemove an API Key')
    print(' -al\t\t\t\t\t\tList all API Keys in database')

    sys.exit(0)


class DBConf:
    REQUESTS = {
        'y_get_active':		'SELECT active FROM yubikeys WHERE nickname = ?',
        'y_set_active':		'UPDATE yubikeys SET active = ? WHERE nickname = ?',
        'y_delete':		'DELETE FROM yubikeys WHERE nickname = ?',
        'y_count_nickname':	'SELECT count(nickname) FROM yubikeys WHERE nickname = ? OR publicname = ?',
        'y_add':		'INSERT INTO yubikeys VALUES (?, ?, ?, ?, ?, 1, 1, 1)',

        'oath_get_active':	'SELECT active FROM oathtokens WHERE nickname = ?',
        'oath_set_active':	'UPDATE oathtokens SET active = ? WHERE nickname = ?',
        'oath_delete':		'DELETE FROM oathtokens WHERE nickname = ?',
        'oath_count_nickname':	'SELECT count(nickname) FROM oathtokens WHERE nickname = ? OR publicname = ?',
        'oath_add':		'INSERT INTO oathtokens VALUES (?, ?, ?, ?, 1, 1)',

        'api_count_nickname':	'SELECT count(nickname) FROM apikeys WHERE nickname = ?',
        'api_count_nicknames':	'SELECT count(nickname) FROM apikeys',
        'api_get_last_id':	'SELECT id FROM apikeys ORDER BY id DESC LIMIT 1',
        'api_add':		'INSERT INTO apikeys VALUES (?, ?, ?)',
    }

    def __init__(self, filename, verbose=True):
        self.con = sqlite3.connect(filename)
        self.cur = self.con.cursor()
        self.verbose = verbose

    def select(self, req, param):
        self.cur.execute(self.REQUESTS[req], param)
        self.result = self.cur.fetchone()
        return self.result != None

    def update(self, req, param):
        self.con.execute(self.REQUESTS[req], param)
        self.con.commit()

    def log(self, msg):
        if self.verbose:
            print(msg)


class Yubikey(DBConf):
    def add(self, nickname, publicid, secretid, aeskey):
        if len(nickname) > 16 or len(publicid) > 16 or len(secretid) > 12 or len(aeskey) > 32:
            print('Nickname and publicid must be max 16 characters long.')
            print('Secretid must be 12 characters max, aeskey must be 32 characters max.')
            return -1

        self.select('y_count_nickname', [nickname, publicid])
        if self.result[0] != 0:
            self.log('Key is already into database. Delete it before adding the same key!')
            return -1

        t = time.strftime('%Y-%m-%dT%H:%M:%SZ', time.gmtime())
        self.update('y_add', [ nickname, publicid, t, secretid, aeskey ])
        self.log('Key %s added to database.' % nickname)

    def delete(self, nickname):
        if not self.select('y_get_active', [ nickname ]):
            self.log('Key not found.')
            return -1

        self.update('y_delete', [ nickname ])
        self.log('Key %s deleted.' % nickname)

    def disable(self, nickname):
        if not self.select('y_get_active', [ nickname ]):
            self.log('Key not found.')
            return -1

        active = self.result[0]
        if not active:
            self.log('Key is already disabled.')
            return 0

        self.update('y_set_active', [ 0, nickname ])
        self.log('Key %s disabled.' % nickname)

    def enable(self, nickname):
        if not self.select('y_get_active', [ nickname ]):
            self.log('Key not found.')
            return -1

        active = self.result[0]
        if active:
            self.log('Key is already enabled.')
            return 0

        self.update('y_set_active', [ 1, nickname ])
        self.log('Key %s enabled.' % nickname)

    def list(self):
        self.cur.execute('SELECT count(nickname) FROM yubikeys')
        rowcount = self.cur.fetchone()
        self.log('%d keys into database:' % rowcount[0])

        users = []
        if rowcount[0]:
            self.cur.execute('SELECT nickname, publicname, active FROM yubikeys')
            self.log('[Nickname]\t\t>> [PublicID]\t\t>> [Active]')
            for nickname, publicname, active in self.cur:
                self.log('%-23s >> %-20s >> %s ' %  (nickname, publicname, active))
                users.append((nickname, publicname, active))

        return users

class OATH(DBConf):
    def add(self, nickname, publicid, key):
        if len(nickname) > 16 or len(publicid) > 16 or len(key) > 40:
            print('Nickname and publicid must be max 16 characters long.')
            print('Secret key must be 40 characters max.')
            return -1

        self.select('oath_count_nickname', [nickname, publicid])
        if self.result[0] != 0:
            self.log('Key is already into database. Delete it before adding the same key!')
            return -1

        t = time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime())
        self.update('INSERT INTO oathtokens VALUES (?, ?, ?, ?, 1, 1)', [nickname, publicid, t, key])
        self.log('Key "%s" added to database.' % key)

    def disable(self, nickname):
        if not self.select('oath_get_active', [ nickname ]):
            self.log('Key not found.')
            return -1

        active = self.result[0]
        if not active:
            self.log('Key is already disabled.')
            return 0

        self.update('oath_set_active', [ 0, nickname ])
        self.log('Key %s disabled.' % nickname)

    def enable(self, nickname):
        if not self.select('oath_get_active', [ nickname ]):
            self.log('Key not found.')
            return -1

        active = self.result[0]
        if active:
            self.log('Key is already enabled.')
            return 0

        self.update('oath_set_active', [ 1, nickname ])
        self.log('Key %s enabled.' % nickname)

    def delete(self, nickname):
        if not self.select('oath_get_active', [ nickname ]):
            self.log('Key not found.')
            return -1

        self.update('y_delete', [ nickname ])
        self.log('Key %s deleted.' % nickname)

    def list(self):
        cur.execute('SELECT count(nickname) FROM oathtokens')
        rowcount = cur.fetchone()
        self.log(" %d keys into database:" % rowcount[0])

        if rowcount[0] != 0:
            cur.execute('SELECT nickname, publicname, active FROM oathtokens')
            self.log('[Nickname]\t\t>> [PublicID]')
            for (nickname, publicname,active) in cur:
                self.log('%-23s >> %-21s >> %s ' % (nickname, publicname, active))

class API(DBConf):
    def add(self, nickname):
        self.select('api_count_nickname', [nickname])
        if self.result[0] != 0:
            self.log('API Key for this nickname is already present. Remove it or choose another one.')
            return -1

        if not self.select('api_get_last_id', []):
            id = 1
        else:
            lastid = self.result[0]
            id = lastid + 1

        api_key = base64.b64encode(randomChars(20))
        self.update('api_add', [nickname, api_key, id])
        self.log('New API Key for %s: %s' % (nickname, api_key))
        self.log('Your API Key ID is: %d' % id)

    def delete(self, nickname):
        self.select('api_count_nickname', [nickname])
        if self.result[0] == 0:
            self.log("API Key for this nickname doesn't exists!")
            return -1

        self.cur.execute('DELETE FROM apikeys WHERE nickname = ?', [nickname])
        self.con.commit()
        self.log('API Key for %s has been deleted.' % nickname)

    def list(self):
        self.select('api_count_nicknames', [])
        rowcount = self.result[0]
        self.log('%d keys into database:' % rowcount)

        keys = []
        if rowcount != 0:
            self.cur.execute('SELECT id, nickname, secret FROM apikeys')
            self.log('[Id]\t>> [Keyname]\t\t>> [Secret]')
            for id, nickname, secret in self.cur:
                self.log('%-7d >> %-20s >> %s' % (id, nickname, secret))
                keys.append((id, nickname, secret))
        else:
            self.log('No keys in database')
        return keys


def main():
    options = {
        '-ya': (4, Yubikey, 'add'),
        '-yk': (1, Yubikey, 'delete'),
        '-yd': (1, Yubikey, 'disable'),
        '-ye': (1, Yubikey, 'enable'),
        '-yl': (0, Yubikey, 'list'),

        '-ha': (3, OATH, 'add'),
        '-hk': (1, OATH, 'delete'),
        '-hd': (1, OATH, 'disable'),
        '-he': (1, OATH, 'enable'),
        '-hl': (0, OATH, 'list'),

        '-aa': (1, API, 'add'),
        '-ak': (1, API, 'delete'),
        '-al': (0, API, 'list'),
    }

    argv = sys.argv[1:]
    if len(argv) == 0:
        usage()

    if argv[0] not in options:
        usage()

    n, klass, fname = options[argv[0]]
    if len(argv[1:]) != n + 1:
        usage()

    filename = argv[-1]
    if not os.path.exists(filename):
        print('SQLite database "%s" doesn\'t exist' % filename)
        sys.exit(1)

    args = argv[1:-1]
    function = getattr(klass, fname)
    db = klass(filename)
    function(db, *args)

if __name__ == '__main__':
    main()



#!/usr/bin/env python3

import http.server
import socketserver
import base64
import hashlib
import hmac
import optparse
import os
import re
import signal
import socket
import sys
import threading
import time
import urllib.request, urllib.parse, urllib.error
import urllib.parse

from yubikeyedup import yubistatus
from yubikeyedup import validate
from yubikeyedup import html
from yubikeyedup.sql import *


class YubiServeHandler:
    def __init__(self, sql_connection, params, vclass):
        self.sql = SQL(sql_connection)
        self.params = params
        self.vclass = vclass

    def sign_message(self, answer, api_key):
        data = [ '%s=%s' % (k, v) for (k, v) in answer.items() ]
        data.sort()
        data = '&'.join(data)

        otp_hmac = hmac.new(api_key, data.encode('utf-8'), hashlib.sha1)
        otp_hmac = base64.b64encode(otp_hmac.digest()).decode('utf-8')

        return otp_hmac

    def build_answer(self, status, answer, api_key=bytes()):
        answer['status'] = status
        answer['h'] = self.sign_message(answer, api_key)

        data = '\r\n'.join([ '%s=%s' % (k, v) for (k, v) in answer.items() ])
        data += '\r\n'

        return data

    def do_validate(self):
        answer = { 't': time.strftime("%Y-%m-%dT%H:%M:%S"), 'otp': '' }

        # API id and OTP are required
        if 'id' not in self.params or 'otp' not in self.params:
            return self.build_answer(yubistatus.MISSING_PARAMETER, answer)

        # ensure API id is valid
        if not self.sql.select('get_api_secret', [self.params['id']]):
            return self.build_answer(yubistatus.NO_SUCH_CLIENT, answer)

        api_key = base64.b64decode(self.sql.result[0])

        # do token validation
        vclass = self.vclass(self.sql)
        status = vclass.set_params(self.params, answer)
        if status == yubistatus.OK:
            status = vclass.validate()
        return self.build_answer(status, answer, api_key)


class YubiHTTPServer(http.server.BaseHTTPRequestHandler):
    __base = http.server.BaseHTTPRequestHandler
    __base_handle = __base.handle

    server_version = 'YubiKeyedUp/1.0'
    vclasses = {
        '/wsapi/2.0/verify': validate.Yubico,
        '/wsapi/2.0/oathverify': validate.OATH,
    }

    PARAM_REGEXP = {
        'id': '[0-9]{1,3}',
        'otp': '[cbdefghijklnrtuv]{0,16}[cbdefghijklnrtuv]{32}',
        'publicid': '[cbdefghijklnrtuv]{0,16}',
        'nonce': '[a-zA-Z0-9]{16,40}',
    }

    def __init__(self, request, client_address, server):
        global sqlite_db
        self.sql_connection = connect_to_db(sqlite_db)
        return http.server.BaseHTTPRequestHandler.__init__(self, request, client_address, server)

    def setup(self):
        self.connection = self.request
        self.rfile = self.request.makefile( 'rb', self.rbufsize)
        self.wfile = self.request.makefile( 'wb', self.wbufsize)

    def getToDict(self, qs):
        dict = {}
        for singleValue in qs.split('&'):
            if not '=' in singleValue:
                continue
            key, value = singleValue.split('=', 1)
            value = urllib.parse.unquote_plus(value)
            if key in self.PARAM_REGEXP and re.match(self.PARAM_REGEXP[key], value):
                dict[key] = value
        return dict

    def do_GET(self):
        url = urllib.parse.urlparse(self.path, 'http')

        if url.path in self.vclasses:
            params = self.getToDict(url.query)
            vclass = self.vclasses[url.path]
            handler = YubiServeHandler(self.sql_connection, params, vclass)
            data = handler.do_validate()
            content_type = 'text/plain'
        else:
            data = index = html.index
            content_type = 'text/html'

        self.send_response(200)
        self.send_header('Content-type', content_type)
        self.end_headers()
        self.wfile.write(data.encode('utf-8'))


class ThreadingHTTPServer(socketserver.ThreadingMixIn, http.server.HTTPServer):
    pass


def stop_signal_handler(signum, frame):
    yubiserveHTTP.shutdown()
    sys.exit(0)


def main():
    parser = optparse.OptionParser('Usage: %prog [options]')
    parser.add_option('-d', '--db', default='./yubikeys.sqlite3', dest='db')
    parser.add_option('-a', '--address', default='0.0.0.0', dest='host')
    parser.add_option('-p', '--port', default='8000', dest='port')
    (options, args) = parser.parse_args()
    global sqlite_db
    sqlite_db = options.db

    yubiserveHTTP = ThreadingHTTPServer((options.host, int(options.port)), YubiHTTPServer)

    signal.signal(signal.SIGINT, stop_signal_handler)

    http_thread = threading.Thread(target=yubiserveHTTP.serve_forever)
    http_thread.setDaemon(True)
    http_thread.start()

    signal.pause()


if __name__ == '__main__':
    main()



import sqlite3


class SQL:
    REQUESTS = {
        'yubico_get_key':	 'SELECT aeskey, internalname, counter, time FROM yubikeys WHERE publicname = ? AND active = 1',
        'yubico_update_counter': 'UPDATE yubikeys SET counter = ?, time = ? WHERE publicname = ?',

        'oath_get_token':	 'SELECT counter, secret FROM oathtokens WHERE publicname = ? AND active = 1',
        'oath_update_counter':	 'UPDATE oathtokens SET counter = ? WHERE publicname = ?',

        'get_api_secret':	 'SELECT secret from apikeys WHERE id = ?',
    }

    def __init__(self, con):
        self.con = con
        self.cur = self.con.cursor()

    def select(self, req, param):
        self.cur.execute(self.REQUESTS[req], param)
        self.result = self.cur.fetchone()
        return self.result != None

    def update(self, req, param):
        self.con.execute(self.REQUESTS[req], param)
        self.con.commit()

def connect_to_db(filename):
    return sqlite3.connect(filename, check_same_thread=False)



import re

from Cryptodome.Cipher import AES

from yubikeyedup.sql import *
from yubikeyedup import yubistatus


class Validate:
    def __init__(self, sql):
        self.sql = sql


class Yubico(Validate):
    # sorry for this one-liner
    modhex = ''.join(dict([ ('cbdefghijklnrtuv'[i], '0123456789abcdef'[i]) for i in range(16)] ).get(chr(j), '?') for j in range(256))

    def set_params(self, params, answer):
        if 'nonce' not in params:
            return yubistatus.MISSING_PARAMETER

        answer['otp'] = params['otp']
        answer['nonce'] = params['nonce']
        answer['sl'] = '100'

        self.otp = params['otp']

        return yubistatus.OK

    def modhexdecode(self, string):
        return bytes.fromhex(string.translate(self.modhex))

    def CRC(self, data):
        crc = 0xffff
        for b in data:
            crc ^= (b & 0xff)
            for j in range(0, 8):
                n = crc & 1
                crc >>= 1
                if n != 0:
                    crc ^= 0x8408
        return crc

    def validate(self):
        match = re.match('([cbdefghijklnrtuv]{0,16})([cbdefghijklnrtuv]{32})', self.otp)
        if not match:
            # this should not happen because otp matches YubiHTTPServer.PARAM_REGEXP
            return yubistatus.BACKEND_ERROR

        userid, token = match.groups()

        if not self.sql.select('yubico_get_key', [userid]):
            return yubistatus.BAD_OTP
        aeskey, internalname, counter, time = self.sql.result

        aes = AES.new(bytes.fromhex(aeskey), AES.MODE_ECB)
        plaintext = aes.decrypt(self.modhexdecode(token)).hex()

        if internalname != plaintext[:12]:
            return yubistatus.BAD_OTP

        # if self.CRC(plaintext[:32].decode('hex')) != 0xf0b8:
        if self.CRC(bytes.fromhex(plaintext[:32])) != 0xf0b8:
            return yubistatus.BAD_OTP

        internalcounter = int(plaintext[14:16] + plaintext[12:14] + plaintext[22:24], 16)
        if counter >= internalcounter:
            return yubistatus.REPLAYED_OTP

        timestamp = int(plaintext[20:22] + plaintext[18:20] + plaintext[16:18], 16)
        if time >= timestamp and (counter >> 8) == (internalcounter >> 8):
            return yubistatus.BAD_OTP

        self.sql.update('yubico_update_counter', [internalcounter, timestamp, userid])

        return yubistatus.OK


class OATH(Validate):
    def set_params(self, params, answer):
        if len(otp) in [ 18, 20 ]:
            publicid = otp[0:12]
            oath = params['otp'][12:]
        elif len(otp) in [ 6, 8 ]:
            if 'publicid' not in params:
                return yubistatus.MISSING_PARAMETER
            publicid = params['publicid']
            oath = params['otp']
        else:
            return yubistatus.BAD_OTP

        answer['otp'] = params['otp']

        self.oath = oath
        self.publicid = publicid

        return yubistatus.OK

    def test_hotp(self, key, counter, digits=6):
        counter = str(counter).rjust(16, '0').decode('hex')
        hs = hmac.new(key, counter, hashlib.sha1).digest()
        offset = ord(hs[19]) & 0xF
        bin_code = int((chr(ord(hs[offset]) & 0x7F) + hs[offset+1:offset+4]).encode('hex'), 16)
        return str(bin_code)[-digits:]

    def validate(self):
        # XXX: TODO, it hasn't been tested
        return yubistatus.BACKEND_ERROR

        if len(self.oath) % 2 != 0:
            return yubistatus.BAD_OTP

        if not self.sql.select('oath_get_token', [publicid]):
            return yubistatus.BAD_OTP

        actualcounter, key = self.sql.result
        key = key.decode('hex')
        for counter in range(actualcounter + 1, actualcounter + 256):
            if self.oath == self.test_hotp(key, counter, len(self.oath)):
                self.sql.update('yubico_update_counter', [str(counter), self.publicid])
                return yubistatus.OK

        return yubistatus.BAD_OTP



OK = 'OK'
BAD_OTP = 'BAD_OTP'
REPLAYED_OTP = 'REPLAYED_OTP'
BAD_SIGNATURE = 'BAD_SIGNATURE'
MISSING_PARAMETER = 'MISSING_PARAMETER'
NO_SUCH_CLIENT = 'NO_SUCH_CLIENT'
OPERATION_NOT_ALLOWED = 'OPERATION_NOT_ALLOWED'
BACKEND_ERROR = 'BACKEND_ERROR'
NOT_ENOUGH_ANSWERS = 'NOT_ENOUGH_ANSWERS'
REPLAYED_REQUEST = 'REPLAYED_REQUEST'



#!/usr/bin/env python3

import base64
import hashlib
import hmac
import http.client
import os
import re
import subprocess
import time
import unittest

import dbconf
import dbcreate

testuser = "nelg"
testserver = "localhost:8001"
yubicotesturl = "/wsapi/2.0/verify"
valid_otp = 'hihrhghufvfibbbekurednelnklnulclbiubvjrenlii'
nonce = 'nILSu3qRwoDldXjgU1'



class YubiserveTestCase(unittest.TestCase):
    def setUp(self):
        self.tearDown()
        return yubikey.add(testuser, 'hihrhghufvfi', '676f6e656c67', '89eb6d3d930077b427a88760db0fc375')

    def tearDown(self):
        return yubikey.delete(testuser)


class YubikeyTestCase(YubiserveTestCase):
    def testDbAPIkey(self):
        apikeys = api.list()
        self.assertTrue(len(apikeys) > 0, msg="No API keys in DB")

    def testDbconf(self):
        users = yubikey.list()
        nicknames = [ (nickname, publicid) for nickname, publicid, _ in users ]
        self.assertTrue((testuser, 'hihrhghufvfi') in nicknames)

    def testDisableEnableKey(self):
        yubikey.disable(testuser)
        data = self.curl('?id=1&otp=%s&nonce=%s' % (valid_otp, nonce))
        m = re.search("^status=BAD_OTP", data, re.M)
        self.assertTrue(m, msg="Valid yubikey, but account is disabled")

        yubikey.enable(testuser)
        data = self.curl('?id=1&otp=%s&nonce=%s' % (valid_otp, nonce))
        m = re.search("^status=OK", data, re.M)
        self.assertTrue(m, msg="Valid yubikey not accepted")

    def testSignature(self):
        data = self.curl('?id=1&otp=%s&nonce=%s' % (valid_otp, nonce))
        m = re.search("^status=OK", data, re.M)
        self.assertTrue(m, msg="Valid yubikey not accepted")

        h = re.findall('h=([^\r]+)', data)
        self.assertTrue(len(h) == 1, msg="Invalid hash")
        h = h[0]

        data = data.split('\r\n')
        data.sort()
        data = '&'.join(l for l in data if not l.startswith('h=') and l != '')
        _, _, api_key = [ keys for keys in api.list() if keys[1] == 'test' ][0]
        otp_hmac = hmac.new(base64.b64decode(api_key), data.encode('utf-8'), hashlib.sha1)
        otp_hmac = base64.b64encode(otp_hmac.digest()).decode('utf-8')
        self.assertTrue(otp_hmac == h, msg="Invalid signature")

    def testReplayKey(self):
        data = self.curl('?id=1&otp=%s&nonce=%s' % (valid_otp, nonce))
        m = re.search("^status=OK", data, re.M)
        self.assertTrue(m, msg="Valid yubikey not accepted")

        data = self.curl('?id=1&otp=%s&nonce=%s' % (valid_otp, nonce))
        m = re.search("^status=REPLAYED_OTP", data, re.M)
        self.assertTrue(m, msg="Replayed token should not be accepted")

    def testbadCRC(self):
        invalid_otp = 'hihrhghufvfirvbegrijgdjhjhtgihcehehtcrgbrhrb'
        data = self.curl('?id=1&otp=%s&nonce=%s' % (invalid_otp, nonce))
        m = re.search("^status=BAD_OTP", data, re.M)
        self.assertTrue(m, msg="Yubikey with Bad CRC")

    def testInvalidInput(self):
        data = self.curl('?id=1&otp=&&&&&&&&&&&&&&&&&&&&&&&&')
        m = re.search("^status=MISSING_PARAMETER", data, re.M)
        self.assertTrue(m, msg="invalid input should not be accepted, otp not set")

    def testInvalidOTP(self):
        invalid_otp = 'hihrhghufvfibbbek1urednelnklnulclbiubvjrenlii'
        data = self.curl('?id=1&otp=%s&nonce=%s' % (invalid_otp, nonce))
        m = re.search("^status=MISSING_PARAMETER", data, re.M)
        self.assertTrue(m, msg="invalid otp (contains a 1)")

    def testInvalidId(self):
        data = self.curl('?id=100&otp=%s&nonce=%s' % (valid_otp, nonce))
        m = re.search("^status=NO_SUCH_CLIENT", data, re.M)
        self.assertTrue(m, msg="API id shouldn't be accepted")

    def curl(self, params='?id=1&otp=%s' % valid_otp, debug=False, url=yubicotesturl, httpcode=200):
        conn = http.client.HTTPConnection(testserver)
        conn.request('GET', url + params)
        r = conn.getresponse()
        self.assertTrue(r.status == httpcode)
        data = r.read().decode('utf-8')
        conn.close()
        if debug:
            print('Request: %s\n' % (url+params))
            print(data)
        return data


def wait_for_server(testserver=testserver):
    while True:
        try:
            conn = http.client.HTTPConnection(testserver)
            conn.request('GET', "/")
            conn.close()
            return
        except IOError:
            time.sleep(0.1)


if __name__ == '__main__':
    filename = '/tmp/yubiservetest.sqlite3'

    try:
        os.unlink(filename)
    except OSError:
        pass

    # create database
    dbcreate.create_db(filename)

    # add an API key
    api = dbconf.API(filename, verbose=False)
    api.add('test')

    yubikey = dbconf.Yubikey(filename, verbose=False)

    p = subprocess.Popen([ './src/yubiserve.py', '--db', filename, '--port=8001' ], stderr=open('/dev/null', 'w'))
    wait_for_server()

    try:
        suite = unittest.TestLoader().loadTestsFromTestCase(YubikeyTestCase)
        unittest.TextTestRunner().run(suite)
    finally:
        api.delete('test')
        os.unlink(filename)

        p.kill()
